# Copyright 2011 Jason Gerecke
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.

"""PyGTK backend for dealing with sprites.

This module contains all the code necessary to make PyGTK work according
to our needs. Its 'main' method is responsible for setting up the library,
and the classes contained implement the GUI portions of the program.

The PyGTK library is available at http://www.pygtk.org/ and must be installed
for this module to load.
"""

import math
import random
import threading
import time
import cairo

try:
	import pygtk
	pygtk.require("2.0")
except:
	pass
try:
	import gtk
	import gtk.glade
	import gobject
except:
	sys.exit(1)

from . import task_collection
from . import sprite


thread = task_collection.TaskManager(lambda fn: gobject.idle_add(fn))
supports_alpha = False

def main(loader):
	"""Initialize the PyGTK library."""
	gtk.gdk.threads_init()
	set_colormap()
	SpriteManagerWindow(loader)
	gtk.main()

def set_colormap():
	"""Switch to a colormap which supports alpha blending if possible.
	
	Based on http://mikehearn.wordpress.com/2006/03/26/gtk-windows-with-alpha-channels/
	"""
	global supports_alpha
	screen = gtk.gdk.Screen()
	colormap = screen.get_rgba_colormap()

	if colormap == None:
		print 'Your screen does not support alpha channels!'
		supports_alpha = False
		colormap = screen.get_rgb_colormap()
	else:
		print 'Your screen supports alpha channels!'
		supports_alpha = True
	screen.set_default_colormap(colormap)

def build_menu(menu):
	"""Build a gtk.Menu object from a description of the menu.
	
	The menu datastructure is a list of tuples, with the first
	element of the tuple being the name of the menu item and the
	second element either a function or a submenu. A function
	is a tuple whose first element is the function itself, and the
	second element the arguments the function should be run with.
	
	menu = [ \
		("Help", (helpdialog, [arg1, arg2])), \
		("Recent", [ \
			("Item 1", (open, [filename])), \
			("Item 2", (open, [filename])) \
		]), \
		("Exit", (quit, [])) \
	]
	"""
	if isinstance(menu, list):
		gtkmenu = gtk.Menu()
		for child in menu:
			gtkmenu.append(build_menu(child))
		return gtkmenu

	elif isinstance(menu, tuple) and len(menu) == 2:
		name, data = menu
		menuitem = gtk.MenuItem(name)

		if isinstance(data, list):
			menuitem.set_submenu(build_menu(data))

		elif isinstance(data, tuple):
			fn, args = data
			def f(object, args):
				thread.add(fn, args)

			menuitem.connect("activate", f, args)
		else:
			raise Exception("Malformed menu.")

		return menuitem

	else:
		raise Exception("Malformed menu.")

class SpriteManagerWindow(gtk.Window):
	"""A window which lists sprites and allows for their creation."""

	def __init__(self, loader):
		self.loader = loader
		self.controllers = []

		gtk.Window.__init__(self, gtk.WINDOW_TOPLEVEL)
		self.set_title("Desktop Ponies")
		self.connect("destroy", gtk.main_quit)
		self.set_size_request(250,300)

		screen = gtk.gdk.Screen(self.get_window())
		self._universe = sprite.Universe(0, 0, screen.get_width(), screen.get_height())

		category_gui = {'- All -': []}
		for s in loader.sprites:
			category_gui['- All -'].append(s)
			if len(s['categories']) == 0:
				continue
			for category in s['categories']:
				key = category.title()
				if not key in category_gui:
					category_gui[key] = []
				category_gui[key].append(s)

		category_store = gtk.ListStore(str, object)
		for key in sorted(category_gui.keys()):
			category_store.append(row=(key, category_gui[key]))

		main_vbox = gtk.VBox(homogeneous=False)
		self.add(main_vbox)

		categories = gtk.ComboBox(model=category_store)
		cell = gtk.CellRendererText()
		categories.pack_start(cell, True)
		categories.add_attribute(cell, 'text', 0)
		main_vbox.pack_start(categories, expand=False, fill=False)

		scroll = gtk.ScrolledWindow()
		scroll.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_ALWAYS)
		viewport = gtk.Viewport()
		scroll.add(viewport)
		main_vbox.add(scroll)

		self.show_all()

		categories.connect("changed", self.category_dropdown_changed, viewport)
		categories.set_active(0)

		self._statusicon = gtk.status_icon_new_from_file("resources/icons/twilight.png")
		self._statusicon.connect('activate', self.status_clicked)
		self._statusicon.connect('popup-menu', self.status_popup)
		self._statusicon.set_tooltip(self.get_title())

		self._statusmenu = gtk.Menu()
		mi = gtk.MenuItem(label="Show/Hide Main Window")
		self._statusmenu.append(mi)
		mi.connect("activate", self.status_clicked)
		mi.show()
		mi = gtk.MenuItem(label="Quit")
		self._statusmenu.append(mi)
		mi.connect("activate", gtk.main_quit)
		mi.show()

	def build_selector(self, sprite):
		def g(widget, data=None):
			i = widget.get_value_as_int()
			z = filter(lambda c: c.model.name == data, self.controllers)
			for x in range(len(z), i):
				controller = self.loader.spawn(data, self._universe)
				controller.start()
				self.controllers.append(controller)
			if len(z) > i:
				for c in z[i-len(z):]:
					c.stop()
					self._universe.sprites.remove(c.model)
					self.controllers.remove(c)

		spin = gtk.SpinButton()
		spin.set_increments(1,1)
		spin.set_range(0,99)

		z = filter(lambda x: x.name == sprite['name'], self._universe.sprites)
		spin.set_value(len(z))

		spin.connect("value-changed", g, sprite['name'])
		label = gtk.Label(sprite['name'])
		label.set_alignment(0.0, 0.5)
		hbox = gtk.HBox()
		hbox.pack_start(spin, expand=False, fill=False)
		hbox.add(label)
		return hbox

	def status_clicked(self, status):
		if self.get_visible():
			self.hide_all()
		else:
			self.show_all()
			self.present()
			
	def status_popup(self, status_icon, button, activate_time):
		self._statusmenu.popup(None, None, gtk.status_icon_position_menu, button, activate_time, status_icon)

	def category_dropdown_changed(self, combobox, viewport):
		child = viewport.get_child()
		if child is not None:
			viewport.remove(child)

		model = combobox.get_model()
		active = combobox.get_active()

		vbox = gtk.VBox()
		for sprite in model[active][1]:
			vbox.add(self.build_selector(sprite))

		viewport.add(vbox)
		viewport.show_all()


class TextBubble(gtk.Window):
	"""A Window which serves as a comic-esque speech bubble."""

	def __init__(self, text):
		gtk.Window.__init__(self, gtk.WINDOW_POPUP)
		self.set_type_hint(gtk.gdk.WINDOW_TYPE_HINT_TOOLTIP)
		self.add(gtk.Label(text))

	@staticmethod
	def say(text, location, timeout=5000):
		x, y = location
		bubble = TextBubble(text)
		bubble.move(int(x), int(y))
		bubble.show_all()
		gobject.timeout_add(timeout, bubble.destroy)


class SpriteWindow(gtk.Window):
	"""A window which draws a sprite to the screen.
	
	The SpriteWindow class creates an undecorated "transparent" window
	which moves around as the sprite moves, and is updated to reflect
	the current sprite image. Since the user is not expected to interact
	with sprites like a normal window, these windows should not appear
	in the desktop environment's window list, ALT+TAB list, workspace
	switcher, or other places that would belie the fact that they are
	*windows* and not graphics blitted directly to the screen.
	"""

	def __init__(self):
		gtk.Window.__init__(self, gtk.WINDOW_POPUP)

		self.connect('expose-event', self.expose)
		self.connect('enter-notify-event', self.mousein)
		self.connect('button-press-event', self._button_press)
		self.connect('button-release-event', self._button_release)
		self.connect('focus-in-event', self._focus_in)
		self.connect('motion-notify-event', self._motion_notify)
		events = gtk.gdk.BUTTON_PRESS_MASK   | \
		         gtk.gdk.BUTTON_RELEASE_MASK | \
		         gtk.gdk.FOCUS_CHANGE_MASK   | \
		         gtk.gdk.POINTER_MOTION_MASK
		self.set_events(events)

		self.set_app_paintable(True)

		self.image_iter = None
		self._animating = None
		self._dragging = None
		self._maskupdate = False
		self._pixbuf = None

	def set_image(self, filename):
		def iterate(filename):
			if filename == self._animating:
				self._set_background()
				thread.add(iterate, [filename], min_wait = self.image_iter.get_delay_time() / 1000.0)

		if filename is None:
			self._animating = filename
			return
		if filename != self._animating:
			self._animating = filename
			animation = gtk.gdk.PixbufAnimation(filename)
			self.set_size_request(animation.get_width(), animation.get_height())
			self.image_iter = animation.get_iter()
			thread.add(iterate, [filename])

	def _set_background(self):
		self.image_iter.advance()
		self._pixbuf = self.image_iter.get_pixbuf()
		self._set_mask()
		if not self.get_visible():
			self.show()
		else:
			self.expose(self, gtk.gdk.Event(gtk.gdk.EXPOSE))

	def _set_mask(self):
		gdkWindow = self.get_window()
		if gdkWindow is None:
			return

		if not supports_alpha:
			pixmap, mask = self._pixbuf.render_pixmap_and_mask()
			gdkWindow.shape_combine_mask(mask, 0, 0)
		elif self._maskupdate:
			px, py, f = gdkWindow.get_pointer()
			w, h      = self.get_size()

			if px < 0 or py < 0 or px > w or py > h:
				# The pointer is outside the maximum window extents, so clear the mask
				self._maskupdate = False
				gdkWindow.input_shape_combine_region(gtk.gdk.region_rectangle((0, 0, w, h)), 0, 0)
			else:
				pixmap, mask = self._pixbuf.render_pixmap_and_mask()
				gdkWindow.input_shape_combine_mask(mask, 0, 0)

	def _button_press(self, widget, event):
		if event.button == 1:
			self._dragging = (event.x, event.y)

		if event.button == 3:
			self._gtkmenu.show_all()
			self._gtkmenu.popup(None, None, None, event.button, event.get_time())
			return True

	def _button_release(self, widget, event):
		if event.button == 1:
			self._dragging = None

	def _motion_notify(self, widget, event):
		if self._dragging is None:
			return False
		rootwin = widget.get_screen().get_root_window()
		px, py, mods = rootwin.get_pointer()
		rx, ry = self._dragging
		self.move(int(px - rx),int(py - ry))

	def _focus_in(self, widget, event):
		if self._dragging:
			# Dragging has stopped, so reset the sprite position.
			self._dragging = False

	def expose(self, widget, event):
		"""Handle GTK's window expose event.
		
		Renders the current frame onto a transparent background.
		Based on http://mikehearn.wordpress.com/2006/03/26/gtk-windows-with-alpha-channels/
		"""
		if not self.get_visible():
			return False

		cr = widget.window.cairo_create()

		# Draw the background
		if supports_alpha == True:
			cr.set_source_rgba(1.0, 1.0, 1.0, 0.0) # Transparent
		else:
			cr.set_source_rgb(1.0, 1.0, 1.0) # Opaque white
		cr.set_operator(cairo.OPERATOR_SOURCE)
		cr.fill()

		# Draw the current frame
		cr.set_source_pixbuf(self._pixbuf, 0, 0)
		cr.paint()

		return False

	def mousein(self, widget, data):
		if not self._maskupdate:
			self._maskupdate = True
			self._set_mask()

	def set_context_menu(self, arg):
		self._gtkmenu = build_menu(arg)

