# Copyright 2011 Jason Gerecke
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.

"""Basic classes for describing sprites in the model-view-controller pattern.

Sprites run the gamut from dumb static images, through "intelligent"
objects that have a mind of their own, to things which can control or
be controlled by outside forces (such as the user or other sprites).

To reduce the amount of duplicated code, we use the MVC pattern to break
a sprite into an object containing attributes (a Model), one or more
objects capable of modifying those attributes (Controllers), and one or
more objects capable of displaying the internal state (Views).

The classes in this module are meant to be subclassed and do not combine
to provide an interesting sprite on their own.
"""

import random
import time
import math

from ..engine import task_collection


class Model:
	"""Contain the internal state of a sprite.
	
	A Model defines attributes about a sprite such as the image to
	display on screen, its location, motion vector, etc. This Model
	class should provide enough data for basic sprites.
	"""

	def __init__(self, universe):
		self.universe  = universe
		self._location = self.universe.random_location()
		self._location_t = 0
		self._speed    = 0
		self._angle    = 0
		self._images   = [(None, None)]

		self.universe.add(self)

	def get_location(self):
		deltaT = time.time() - self._location_t
		x, y   = self._location

		# Remember kids, the screen's y-axis is inverted!
		x      = x + (math.cos(math.radians(self._angle)) * self._speed * deltaT * 25)
		y      = y - (math.sin(math.radians(self._angle)) * self._speed * deltaT * 25)

		return (x,y)

	def set_images(self, images):
		"""Set the images associated with angles of travel.

		Oftentimes a sprite will present different images when it
		is moving or facing different directions. This method
		allows the assignment of these images so that the proper
		one may be chosen as needed.
		
		This method expects a list of tuples with the following
		format:
		
		[ (angle_start, filename), (angle_start, filename) ... ]
		
		The appropriate file is chosen by looking for the first
		tuple with an angle_start greater-than or equal to the
		angle of the sprite's motion.
		"""
		self._images = images

	def set_location(self, location):
		"""Teleport the Model to a new location.
		
		This method only adjusts the location of the Model. It
		does not change the angle of travel or speed.
		"""
		self._location_t = time.time()
		self._location   = location

	def set_motion(self, location=None, speed=None, angle=None):
		"""Assign all motion-related variables for the Model.
		
		Changes the location, angle of travel, and speed of the
		Model. Since the angle may change, the list of images is
		consulted to choose a new appropriate image to display.

		If an argument is left unspecified (or assigned to None)
		the current value will be left untouched.
		"""
		if location is not None:
			self.set_location(location)
		else:
			self.set_location(self.get_location())
		if speed    is not None: self._speed = speed
		if angle    is not None: self._angle = angle % 360.0

		filename = self._images[0][1]
		for a, f in self._images:
			if a < self._angle:
				filename = f
			elif a == angle:
				filename = f
				break
			else:
				break

		self._view.set_move(self._speed != 0)
		self._view.set_image(filename)

	def set_view(self, view):
		self._view = view

	def run(self):
		return


class View:
	"""Display the internal state of a sprite's Model.
	
	The View class is capable of producing a particular rendering
	of the Model it is associated with. GUI views will likely draw
	the model on-screen at its location. This particular class,
	however, is bound to the command-line and will only print out
	the public properties of the Model.
	"""

	def __init__(self, model):
		self.model = model

	def stop(self):
		pass

	def update(self):
		"""Update the view of the model to reflect the latest info.
		
		Subclasses should override this method to perform something
		actually useful. For example, a View which draws the sprite
		onto the desktop could read the Model's location and image.
		This brain-dead function simply dumps the data stored by
		the model -- not a particularly useful view of the Model.
		"""
		print self.model.__dict__


class WindowView(View):
	"""Display the sprite's Model with a Window drawn by some backend.
	
	This kind of View uses a Sprite window class provided by an arbitrary
	backend (e.g. pygtk) to display the sprite on-screen. The Sprite
	object has its background and position updated as necessary to stay
	in sync with the internal state of the Model.
	"""

	tasks = task_collection.TaskManager(sleep=0.01)

	def __init__(self, backend, controller):
		View.__init__(self, controller.model)

		self._backend    = backend
		self._controller = controller
		self._window     = backend.SpriteWindow()

		self._moving = False
		self._dragging = False

		x, y = self.model.get_location()
		self._window.move(int(x), int(y))

	def stop(self):
		self.set_move(False)
		self._window.set_image(None)
		self._backend.thread.add(self._window.destroy, [], priority=1)

	def update(self):
		self.set_move(self._controller.model._speed != 0)
		self.set_image(self._controller.model.image)

	def set_context_menu(self, menu):
		"""Assign the context menu items and functions to execute.
		
		The context menu datastructure is a list of tuples, with
		the first element of the tuple being the name of the menu
		item, and the second element either a function or a submenu.
		A function is a tuple whose first element is the function
		itself, and the second element the arguments the function
		should be run with.
		
		menu = [ \
			("Help", (helpdialog, [arg1, arg2])), \
			("Recent", [ \
				("Item 1", (open, [filename])), \
				("Item 2", (open, [filename])) \
			]), \
			("Exit", (quit, [])) \
		]
		"""
		self._window.set_context_menu(menu)

	def Say(self, text, timeout=5000):
		def makebubble():
			x, y = self._controller.model.get_location()
			x, y = x - 10, y - 10
			self._backend.TextBubble.say(text, (x, y), timeout)

		self._backend.thread.add(makebubble, [])

	def set_move(self, do_move):
		def iterate():
			if self._moving:
				self._update_position()
				if self.model._speed != 0.0:
					WindowView.tasks.add(iterate, [], min_wait = 0.1/self.model._speed)
		self._moving = do_move
		WindowView.tasks.add(iterate, [])

	def set_image(self, filename):
		self._window.set_image(filename)

	def _update_position(self):
		if self._window._dragging is not None:
			self._dragging = True
			return
		elif self._dragging == True:
			self._dragging = False
			self.model.set_location(self._window.get_position())
			return

		if not self._moving:
			return

		# Handle universe edge collisions in an ugly way
		top, bottom = self.model.universe.area
		ux1, uy1 = top
		ux2, uy2 = bottom
		x, y   = self.model.get_location()
		angle  = self.model._angle % 360.0
		w, h   = self._window.get_size()

		modify = 0.0

		if x < ux1 and (angle > 90.0 and angle < 270.0):
			# Off left edge, moving left
			modify += 90.0 - angle
		if x+w > ux2 and (angle < 90.0 or angle > 270.0):
			# Off right edge, moving right
			modify += 90.0 - angle
		if y < uy1 and (angle > 0.0 and angle < 180.0):
			# Off top edge, moving up
			modify += 180.0 - angle
		if y+h > uy2 and (angle > 180.0 and angle < 360.0):
			# Off bottom edge, moving down
			modify += 180.0 - angle

		if modify != 0.0:
			angle += modify*2
			self.model.set_motion(location=self._window.get_position(), angle=angle)
		else:
			self._backend.thread.add(self._window.move, [int(x), int(y)])


class Controller:
	"""Control the internal state of a sprite's Model.
	
	The Controller class is a bare-bones implementation which allows
	for the control of a sprite's image and motion. This is sufficient
	to allow for "control" of the simplest of sprites (e.g. static
	background sprites). No run method is provided since there are no
	brains to this simple controller -- all changes to the image and
	motion will come from the user interacting with the view, or from
	some bit of code (perhaps itself in a thread).
	"""

	def __init__(self, backend, model):
		self.backend  = backend
		self.model    = model
		self.view     = WindowView(backend, self)

		self.model.set_view(self.view)

	def start(self):
		pass

	def stop(self):
		self.view.stop()

	def set_image(self, filename):
		self.model.set_image(filename)

	def set_motion(self, location=None, speed=None, angle=None):
		self.model.set_motion(location, speed, angle)


class Universe:
	"""A Universe is a container representing the logical desktop.
	
	The Universe serves two roles: a logical description of the real-
	estate on which sprites exist, and a container to allow sprites
	to more easily interact with each other.
	
	In its first role, a Universe simply defines a region of desktop
	space that is available for sprites to exist in. Controllers
	should consult the Universe before moving a sprite to ensure that
	the destination is valid.
	
	In its second role, a Universe facilitates the interaction between
	sprites. A Controller may consult the list of sprites which are in
	the Universe. It may use this data to activate behaviors based on
	distance, locate a specific sprite to move toward it, etc.
	"""

	def __init__(self, x1, y1, x2, y2):
		"""Create a new Universe for sprites to inhabit."""
		self.area    = ((x1, y1), (x2, y2))
		self.sprites = []

	def entirely_contains(self, location, size=(0,0)):
		"""Determines if the location lies within the Universe.

		If provided with a size, this function will return true
		only if the entire area is inside the Universe.
		"""
		src_x1, src_y1 = location
		src_x2, src_y2 = map(sum, zip(location, size))
		x1, y1, x2, y2 = self.area

		return src_x1 >= x1 and src_x2 >= x1 and \
		       src_x1 <= x2 and src_x2 <= x2 and \
		       src_y1 >= y1 and src_y2 >= y1 and \
		       src_y1 <= y2 and src_y2 <= y2

	def random_location(self):
		"""Obtain a random location within the Universe."""
		return map(random.uniform, *self.area)

	def add(self, model):
		"""Add a sprite's Model to the Universe.

		This function adds a sprite to the Universe, allowing other
		sprites to be made aware of it. After adding a sprite to
		the Universe, its location should be set to a valid value.
		"""
		self.sprites.append(model)

	def vector_from_to(self, source, target):
		"""Obtain a vector from 'source' to 'target'.

		Provided both sprites are in the same Universe (None
		is returned if not so), returns the vector from
		the source to the target. Note that this vector takes
		into account the fact that the Y axis in screen space
		is inverted! If 'target' is 5 pixels to the right and
		10 pixels above 'source' the returned vector is (5,10)
		*NOT* (5,-10) as the raw pixel locations would suggest.
		"""
		if source not in self.sprites or target not in self.sprites:
			return None

		xs, ys = source.get_location()
		xt, yt = target.get_location()

		# Remember kids, the screen's y-axis is inverted!
		return (xt - xs, ys - yt)

	def distance(self, source, target):
		"""Obtain the linear distance from 'source' to 'target'.

		Provided both sprites are in the same Universe (None
		is returned if not so), returns the linear distance
		from the source to the target.
		"""
		def dotproduct(v1, v2):
			return sum((a*b) for a, b in zip(v1, v2))

		def length(v):
			return math.sqrt(dotproduct(v, v))

		if source not in self.sprites or target not in self.sprites:
			return None
		return length(self.vector_from_to(source, target))

	def angle(self, source, target):
		"""Obtain the angle (in radians) from 'source' to 'target'

		Provided both sprites are in the same Universe (None
		is returned if not so), returns the angle in radians
		of the vector from 'source' to 'target'. Angle is relative
		to the X axis (i.e., 0 is to the right, pi/2 above, pi
		(or negative pi) to the left, and -pi/2 below.
		"""
		if source not in self.sprites or target not in self.sprites:
			return None
		v = self.vector_from_to(source, target)
		return math.atan2(v[1], v[0])

	def search(self, source, targets, distance=float('inf')):
		"""Obtain the subset of 'targets' within 'distance' of 'source'.

		Returns the subset of target sprites with matching name
		that are within the specified distance of the source sprite.
		This is used for processing interactions and 'follow' behaviors.
		"""
		match = []
		for name in targets:
			match.extend( \
			   filter( \
			      lambda x: x.name.lower() == name.lower() and self.distance(source, x) < distance,
			      self.sprites
			   )
			)

		return match

