# Copyright 2011 Jason Gerecke
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.

import math
import random
import threading
import time

import wx
import wx.animate

from . import task_collection
from . import sprite


thread = task_collection.TaskManager(lambda fn: wx.CallAfter(fn))

def main(manager):
	app = wx.PySimpleApp()
	SpriteManagerWindow(manager)
	app.MainLoop()

def build_menu(parent, menu, window):
	"""Build a wx.Menu object from a description of the menu.
	
	The menu datastructure is a list of tuples, with the first
	element of the tuple being the name of the menu item and the
	second element either a function or a submenu. A function
	is a tuple whose first element is the function itself, and the
	second element the arguments the function should be run with.
	
	menu = [ \
		("Help", (helpdialog, [arg1, arg2])), \
		("Recent", [ \
			("Item 1", (open, [filename])), \
			("Item 2", (open, [filename])) \
		]), \
		("Exit", (quit, [])) \
	]
	"""
	if isinstance(menu, list):
		wxmenu = wx.Menu()
		for child in menu:
			wxmenu.AppendItem(build_menu(wxmenu, child, window))
		return wxmenu

	elif isinstance(menu, tuple) and len(menu) == 2:
		name, data = menu
		menuitem = wx.MenuItem(parent, wx.ID_ANY, name)

		if isinstance(data, list):
			menuitem.SetSubMenu(build_menu(None, data, window))

		elif isinstance(data, tuple):
			fn, args = data
			def f(object, args):
				thread.add(fn, args)

			window.Bind(wx.EVT_MENU, lambda x: f(x, args), menuitem)
		else:
			raise Exception("Malformed menu.")

		return menuitem

	else:
		raise Exception("Malformed menu.")


class SpriteManagerWindow(wx.Frame):

	def __init__(self, loader):
		self.loader = loader
		self.controllers = []

		wx.Frame.__init__(self, None, wx.ID_ANY, 'Sprite Manager')
		self.Bind(wx.EVT_CLOSE, self.OnClose)

		self.category_gui = {'- All -': []}
		for s in loader.sprites:
			self.category_gui['- All -'].append(s)
			if len(s['categories']) == 0:
				continue
			for category in s['categories']:
				key = category.title()
				if not key in self.category_gui:
					self.category_gui[key] = []
				self.category_gui[key].append(s)

		vbox = wx.BoxSizer(wx.VERTICAL)
		scrollWin = wx.ScrolledWindow(self, wx.ID_ANY)
		categories = wx.ComboBox(self, wx.ID_ANY, value="- All -", choices=sorted(self.category_gui.keys()), style=wx.CB_READONLY)
		categories.Bind(wx.EVT_COMBOBOX, lambda evt: self.category_dropdown_changed(categories, scrollWin))

		geom = wx.Display().GetClientArea()
		self._universe = sprite.Universe(geom[0], geom[1], geom[0]+geom[2], geom[1]+geom[3])

		scrollWin.SetScrollRate(20,20)
		scrollWin.EnableScrolling(True,True)
		vbox.Add(categories, 0, wx.EXPAND)
		vbox.Add(scrollWin, 1, wx.EXPAND)
		self.SetSizer(vbox)
		
		self.Show()

	def OnClose(self, event):
		self.Destroy()
		exit(0)

	def build_selector(self, sprite, parent):
		def f(widget, data):
			i = widget.GetValue()
			z = filter(lambda c: c.model.name == data, self.controllers)
			for x in range(len(z), i):
				controller = self.loader.spawn(data, self._universe)
				controller.start()
				self.controllers.append(controller)
			if len(z) > i:
				for c in z[i-len(z):]:
					c.stop()
					self._universe.sprites.remove(c.model)
					self.controllers.remove(c)

		def g(x, y): # Hack to stop the lambda from rebinding :)
			x.Bind(wx.EVT_SPINCTRL, lambda event: f(x, y))

		z = filter(lambda x: x.name == sprite['name'], self._universe.sprites)

		hbox = wx.BoxSizer(wx.HORIZONTAL)

		spin = wx.SpinCtrl(parent, wx.ID_ANY)
		spin.SetValue(len(z))
		g(spin, sprite['name'])

		label = wx.StaticText(parent, wx.ID_ANY, sprite['name'])

		hbox.Add(label, 2, wx.EXPAND)
		hbox.Add(spin, 1, wx.EXPAND)
		return hbox

	def category_dropdown_changed(self, combobox, scrollwin):
		children = scrollwin.GetChildren()
		for child in children:
			child.Destroy()

		active = combobox.GetValue()

		vbox = wx.BoxSizer(wx.VERTICAL)
		for sprite in self.category_gui[active]:
			vbox.Add(self.build_selector(sprite, scrollwin), 1, wx.EXPAND)

		scrollwin.SetSizer(vbox)
		vbox.FitInside(scrollwin)
		scrollwin.Show()


class TextBubble(wx.Frame):

	def __init__(self, text):
		wx.Frame.__init__(self, None, wx.ID_ANY, "Dialogue", wx.Point(0,0), wx.Size(100,100),
			0 | wx.SIMPLE_BORDER | wx.FRAME_NO_TASKBAR | wx.STAY_ON_TOP)
		vbox = wx.BoxSizer(wx.VERTICAL)
		vbox.Add(wx.StaticText(self, wx.ID_ANY, text))
		self.SetSizer(vbox)
		vbox.Fit(self)

	@staticmethod
	def say(text, location, timeout=5000):
		x, y = location
		bubble = TextBubble(text)
		bubble.Move((int(x), int(y)))
		bubble.Show()
		thread.add(bubble.Destroy, [], min_wait = timeout/1000.0)


class SpriteWindow(wx.Frame):

	def __init__(self):
		wx.Frame.__init__(self, None, wx.ID_ANY, "Sprite", wx.Point(0,0), wx.Size(100,100),
			0 | wx.FRAME_SHAPED | wx.SIMPLE_BORDER | wx.FRAME_NO_TASKBAR | wx.STAY_ON_TOP)

		self.anim_ctrl = None
		self.Bind(wx.EVT_PAINT, self._OnRepaint)
		self._dragging = None

	def get_position(self):
		return self.GetScreenPositionTuple()

	def get_size(self):
		return self.GetClientSize()

	def move(self, x, y):
		self.Move((x,y))

	def set_context_menu(self, arg):
		self._wxmenu = build_menu(None, arg, self)

	def destroy(self):
		self.Destroy()

	def set_image(self, filename):
		def change_anim(self, filename):
			if self.anim_ctrl == None:
				self.anim_ctrl = wx.animate.AnimationCtrl(self, wx.ID_ANY)
				self.anim_ctrl.Bind(wx.EVT_MOUSE_EVENTS, self._OnMotion)

			self.anim_ctrl.Stop()
			self.anim_ctrl.LoadFile(filename)
			self.SetClientSize(self.anim_ctrl.GetSize())
			self.anim_ctrl.Play()
			self.Show()

		if filename:
			thread.add(change_anim, [self, filename], priority=1)

	def _OnMotion(self, event):
		if event.RightDown():
			self.PopupMenu(self._wxmenu)

		if event.Dragging() and self._dragging == None:
			self._dragging = event.GetPosition()
		elif event.Dragging() and self._dragging != None:
			(ini_x, ini_y) = self._dragging
			(rel_x, rel_y) = event.GetPosition()
			delta_x = rel_x - ini_x
			delta_y = rel_y - ini_y

			(abs_x, abs_y) = wx.GetMousePosition()
			x = abs_x - rel_x + delta_x
			y = abs_y - rel_y + delta_y

			thread.add(self.Move, [(x, y)], priority = -1)
		elif event.LeftUp() and self._dragging != None:
			self._dragging = None

	def _OnRepaint(self, id):
		dc = wx.PaintDC(self)
		self.SetShape(wx.RegionFromBitmap(self.anim_ctrl.GetBackingStore()))

