# Copyright 2011 Jason Gerecke
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.

"""Module for loading INI files from the Desktop Ponies propgram.

This module reads and parses the INI files from Desktop Ponies. The INI
files are poorly documented, and there are a number of cases where they
have some qurik that I don't account for. To help mitigate this, the
loader does validation to try to find out up front when things aren't
as we expect them in the rest of the program.
"""

import csv
import os
import codecs
import re
import ponies

def _raise(e, loudness=0):
	if loudness <= 0:
		return
	elif loudness == 1:
		print e
	else:
		raise e

def _string_preserving(string):
	if not isinstance(string, basestring):
		err = "expected string data, got %s" % (repr(string))
		raise(TypeError(err))
	if len(string) == 0:
		err = "string has zero length"
		raise(ValueError(err))
	return string

def _string(string):
	return _string_preserving(string).lower()

def _int(value):
	value = _string(value)
	if value == '':
		err = "expected int data, got empty string"
		raise ValueError(err)
	return int(value)

def _float(value):
	value = _string(value)
	if value == '':
		err = "expected float data, got empty string"
		raise ValueError(err)
	return float(value)

def _bool(boolean):
	boolean = _string(boolean).upper()
	if   boolean == 'TRUE':  return True
	elif boolean == 'FALSE': return False
	err = "'True' or 'False' expected, got %s" % (repr(boolean))
	raise(ValueError(err))

def _movement_angle(direction):
	x = []
	d = _string(direction).upper()
	if "HORIZONTAL" in d or d == "ALL":
		x.append(0)
		x.append(180)
	if "VERTICAL" in d or d == "ALL":
		x.append(90)
		x.append(270)
	if "DIAGONAL" in d or d == "ALL":
		x.append(45)
		x.append(135)
		x.append(225)
		x.append(315)
	if d == "NONE":
		pass

	if len(x) == 0 and d != 'NONE':
		err = "unknown direction %s" % (repr(direction))
		raise(ValueError(err))
	return x

def _path(path):
	path = _string_preserving(path)
	if os.path.exists(path):
		return path

	dir, name = os.path.split(path)
	names = filter(lambda x: x.lower() == name.lower(), os.listdir(dir))
	if len(names) != 1:
		err = "no such file or directory %s" % (repr(path))
		raise ValueError(err)
	else:
		return os.path.join(dir, names[0])


class Factory():
	"""A class capable of loading Pony objects from INI files.
	
	The PonyLoader is used to locate and parse INI files which define
	ponies,	provide an enumeration of available ponies, and construct
	Pony objects on demand.

	The actual INI file format is poorly documented, and somewhat
	difficult to interpret. This is a best guess at translating its
	data into something usable by the program. There's a good chance
	a few bugs are lurking.
	"""

	def __init__(self, backend, root_dir):
		self.backend      = backend
		self.root_dir     = root_dir
		self.sprites      = self._enumerate_sprites()
		self.interactions = self._get_interaction()

	def _enumerate_sprites(self):
		ini_files = []
		for root, dirs, files in os.walk(self.root_dir):
			for name in files:
				if name.lower() == "pony.ini":
					ini_files.append(os.path.join(root, name))

		return map(lambda x: x.data, sorted(filter(lambda x: x!=None, map(PonyLoader, ini_files)), key=lambda x: x.data['name']))

	def spawn(self, name, universe):
		print "Trying to spawn %s" % name
		ini = filter(lambda x: x['name'].upper()==name.upper(), self.sprites)[0]
		interact = filter(lambda x: x['pony'].upper()==name.upper(), self.interactions)
		model = ponies.Model(universe, ini, interact)
		return ponies.Controller(self.backend, model)

	def _get_interaction(self):
		ini_files = []
		interactions = []
		for root, dirs, files in os.walk(self.root_dir):
			for name in files:
				if name.lower() == "interactions.ini":
					ini_files.append(os.path.join(root, name))

		if len(ini_files) > 0:
			for f in ini_files:
				interactions.extend(Factory.get_interactions(f))
		else:
			interactions = None

		return interactions

	@staticmethod
	def get_interactions(ini_file):
		interactions = []
		try:
			data = list(INIReader(ini_file))
		except Exception,e:
			print e
			return None

		for row in data:
			if len(row) == 0:
				continue

			x = {}
			interactions.append(x)
			for prop in ['name', 'pony', 'probability',
			             'proximity', 'targets',
			             'random_or_all', 'behaviors',
			             'delay']:

				value = row.pop(0)
				if prop == 'probability':
					x[prop] = 0.0 if value == '' else float(value)
				elif prop == 'proximity' or prop == 'delay':
					x[prop] = 0 if value == '' else int(value)
				elif prop == 'random_or_all':
					x[prop] = True if value.upper() == 'TRUE' else False
				else:
					x[prop] = value

		return interactions


class PonyLoader:
	"""A class capable of loading Pony objects from INI files.
	
	The PonyLoader reads and interprets INI files which define ponies.
	The actual INI file format is poorly documented, and somewhat
	difficult to interpret. This is a best guess at translating its
	data into something usable by the program. There's a good chance
	a few bugs are lurking.
	"""

	def __init__(self, ini_file):
		self.ini_file = ini_file
		self.lineno = 0
		self.data = self._load()
		PonyValidator.validate(self.data)

	def _load(self):
		info = {'ini_file'  : self.ini_file,
		        'name'      : "",
		        'categories': [],
		        'speak'     : [],
		        'effect'    : [],
		        'behavior'  : []}

		for row in INIReader(self.ini_file):
			self.lineno += 1
			if len(row) == 0:
				continue

			command = self._read(_string, row.pop(0)).lower()
			if   command == 'name'      : info['name'      ] = self._load_name(row)
			elif command == 'categories': info['categories'].extend( self._load_categories(row) )
			elif command == 'speak'     : info['speak'     ].append( self._load_speak(row)      )
			elif command == 'effect'    : info['effect'    ].append( self._load_effect(row)     )
			elif command == 'behavior'  : info['behavior'  ].append( self._load_behavior(row)   )
			else:
				err = "%s, line %d: encountered unkonwn command %s" % \
				      (self.ini_file, self.lineno, repr(command))
				_raise(Exception(err))

			if len(row) != 0:
				err = "%s, line %d: unprocessed data at end of line %s" % \
			              (self.ini_file, self.lineno, repr(row))
				_raise(Exception(err))

		return info

	def _load_name(self, name_data):
		return self._read(_string_preserving, name_data.pop(0))

	def _load_categories(self, category_data):
		categories = []
		while len(category_data) != 0:
			categories.append(self._read(_string, category_data.pop(0)))
		return categories

	def _load_speak(self, speak_data):
		typegroups = [ \
		   [ ('name',       _string,    None, True), ('text', _string_preserving, None,  True)  ], \
		   [ ('file_sound', self._file, None, False), ('skip', _bool,              False, False) ]  \
		]

		if len(speak_data) == 1:
			speak_data.append(speak_data[0])

		return self._load_groups(typegroups, speak_data)

	def _load_effect(self, effect_data):
		#TODO: Replace with proper types once effect code is in place
		typegroups = [ \
		   [ ('name',               _string,    None, True ), ('behavior',        _string,    None, False), \
		     ('file_image_right',   self._file, None, False), ('file_image_left', self._file, None, False), \
		     ('duration',           _float,     None, False), ('delay',           _float,     None, False), \
		     ('rel_location_right', _string,    None, False), ('center_right',    _string,    None, False), \
		     ('rel_location_left',  _string,    None, False), ('center_left',     _string,    None, False), \
		     ('follows',            _string,    None, False) ] \
		]

		return self._load_groups(typegroups, effect_data)

	def _load_behavior(self, behavior_data):
		typegroups = [ \
		   [ ('name',            _string,    None, True), ('probability',       _float,          None, True), \
		     ('duration_max',    _float,     None, True), ('duration_min',      _float,          None, True), \
		     ('movement_speed',  _float,     None, True), ('file_image_right',  self._file,      None, True), \
		     ('file_image_left', self._file, None, True), ('movement_angle',    _movement_angle, [],   True) ], \
		   [ ('linked_behavior', _string, None, False), ('speaking_start', _string, None,  False), \
		     ('speaking_end',    _string, None, False), ('skip',           _bool,   False, False), \
		     ('x_coord',         _float,  None, False), ('y_coord',        _float,  None,  False) ], \
		   [ ('object_to_follow', _string, None, False) ] \
		]

		d = self._load_groups(typegroups, behavior_data)

		# Replace animations with a format that SpriteWindow understands
		d['file_animations'] = [
			(0,   d['file_image_right']),
			(90,  d['file_image_left']),
			(270, d['file_image_right'])
		]
		del d['file_image_right']
		del d['file_image_left']

		return d

	def _read(self, fn, value):
		try: return fn(value)
		except TypeError, e: _raise("%s, line %s: %s" % (self.ini_file, self.lineno, e))

	def _load_groups(self, typegroups, values):
		d = {}
		for g in typegroups:
			for (key, typ, default, required) in g:
				d.setdefault(key, default)
			if len(values) != 0:
				d.update(self._read_group(g, values))
		return d

	def _read_group(self, typeinfo, values):
		d = {}
		for key, typ, default, required in typeinfo:
			try:
				d[key] = self._read(typ, values.pop(0))
			except ValueError, e:
				d[key] = default
				if required:
					err = "%s, line %s: error getting required field '%s': %s" % (self.ini_file, self.lineno, key, e)
					_raise(TypeError(err))
		return d

	def _file(self, paths):
		files = []
		dir = os.path.dirname(self.ini_file)
		if isinstance(paths, basestring):
			paths = [paths]
		for path in paths:
			try:
				files.append(_path(os.path.join(dir, path)))
			except ValueError, e:
				pass
		if len(files) == 0: raise ValueError("no file in list found")
		else: return files[0]


class INIReader:
	"""Reads and parses INI files from the Desktop Ponies project.

	The INI file format used in Desktop Ponies is a slightly customized
	flavor of CSV which allows the storage of arrays. It is typically
	stored in UTF-16, though UTF-8 and other encodings appear from time
	to time. I don't know of a CSV parser that can properly handle the
	array syntax used in these INI files, so this class unfortunatley
	breaks rule #82 of programming: "Don't write your own CSV parser."

	This class operates on the line level, returning a list of lists
	for each (uncommented, non-blank) line. The caller is expected to
	know how to interpret the line.
	"""

	def _build_regex(self):
		"""Build and return the regex used to split lines.

		For reference, the basic idea of the grammar used is as
		follows. Note that the regex actually used doesn't *quite*
		match, since formal_list would require a recursive definition.
		The actual production rule we use is <formal_list> = {<string>}.

			<line>  = <list>
			<list>  = <datum>(,<datum>)*
			<datum> = <formal_list> | <string>
			<formal_list> = {<list>}
			<string> = <quoted_string> | <unquoted_string>

		We're only interested in accessing the datum elements, so
		the regex we use covers just that portion.
		"""
		return '\s*(?:\{(.*?)\}|"(.*?)"|([^,]*))\s*,?'

	def __init__(self, filename):
		encoding = None
		for trial in ["utf-16", "utf-8-sig", "utf-8"]:
			if UTF8Recoder(open(filename), trial).test():
				encoding = trial
				break
		if encoding == None:
			raise Exception("No suitable decoder for '" + filename + "' found.")

		self._pattern = re.compile(self._build_regex())
		self._reader = UTF8Recoder(open(filename), encoding)

	def __iter__(self):
		return self

	def next(self):
		line = None
		line = self._reader.next().strip()
		if self._do_ignore(line):
			return []
		else:
			return self._parse(line)

	def _do_ignore(self, line):
		return line is None         or \
		       len(line) == 0       or \
		       line.startswith("'")

	def _parse(self, line):
		data = []
		for item in re.findall(self._pattern, line):
			if len(item[0]) > 0:   # Lists in first group
				data.append(self._parse(item[0].strip()))
			elif len(item[1]) > 0: # Quoted strings in second group
				data.append(item[1].strip())
			elif len(item[2]) > 0: # Unquoted strings in third group
				data.append(item[2].strip())
			else:
				data.append('')
		return data[:-1] # Our regex has one extra empty match at the end

class UTF8Recoder:
	"""
	Iterator that reads an encoded stream and reencodes the input to UTF-8
	"""
	def __init__(self, f, encoding):
		self.reader = codecs.getreader(encoding)(f)

	def __iter__(self):
		return self

	def next(self):
		return self.reader.next().encode("utf-8")

	def test(self):
		try:
			for line in self.reader:
				line = line.encode("utf-8")
		except UnicodeError:
			return False

		return True


class PonyValidator:
	"""A class for performing sanity checks on pony data.

	This class' purpose is to validate pony data after being loaded
	from an INI file. This checks that the data is semantically sound
	and fits within the expectations of the program. Note that failing
	the validator does *not* necessarily mean the data is bad! Given
	how poorly the INI files are documented, its just as likely to be
	a problem with my assumptions as it is to be with the input. That
	said, failing the validator is not something to be taken lightly.
	"""

	@staticmethod
	def validate(pony):
		PonyValidator._validate_name(pony)
		PonyValidator._validate_categories(pony)
		PonyValidator._validate_speak(pony)
		PonyValidator._validate_effect(pony)
		PonyValidator._validate_behavior(pony)

	@staticmethod
	def _validate_name(pony):
		ini_file = pony['ini_file']
		if len(pony['name']) == 0:
			err = "%s: Pony has name with zero length." % \
		              (ini_file)
			_raise(Exception(err))

	@staticmethod
	def _validate_categories(pony):
		return # No post-validation necessary

	@staticmethod
	def _validate_speak(pony):
		return # No post-validation necessary

	@staticmethod
	def _validate_effect(pony):
		ini_file = pony['ini_file']
		for datum in pony['effect']:
			z = filter(lambda x: datum['name'] == x['name'], pony['effect'])
			if len(z) != 1:
				err = "%s: Effect %s has been defined %d times" % \
			              (ini_file, repr(datum['name']), len(z))
				_raise(Exception(err))

			z = filter(lambda x: datum['behavior'] == x['name'], pony['behavior'])
			if len(z) == 0 and datum['behavior'] is not None:
				err = "%s: Effect %s references undefined behavior %s" % \
			              (ini_file, repr(datum['name']), repr(datum['behavior']))
				_raise(Exception(err))

			if datum['duration'] < 0:
				err = "%s: Effect %s has a duration of %s; expected >= 0" % \
			              (ini_file, repr(datum['name']), repr(datum['duration']))
				_raise(Exception(err))

			if datum['delay'] < 0:
				err = "%s: Effect %s has a delay of %s; expected >= 0" % \
			              (ini_file, repr(datum['name']), repr(datum['delay']))
				_raise(Exception(err))

	@staticmethod
	def _validate_behavior(pony):
		ini_file = pony['ini_file']

		z = reduce(lambda x,y: x+y, map(lambda x: x['probability'], pony['behavior']))
		if z < 0.99 or z > 1.01:
			err = "%s: Behaviors have total probability of %s; expected 0.99 < p < 1.01" % \
		              (ini_file, z)
			_raise(Exception(err))

		for datum in pony['behavior']:
			z = filter(lambda x: datum['name'] == x['name'], pony['behavior'])
			if len(z) != 1:
				err = "%s: Behavior %s has been defined %d times" % \
			              (ini_file, repr(datum['name']), len(z))
				_raise(Exception(err))

			z = filter(lambda x: datum['linked_behavior'] == x['name'], pony['behavior'])
			if len(z) == 0 and datum['linked_behavior'] is not None:
				err = "%s: Behavior %s linked to undefined behavior %s" % \
			              (ini_file, repr(datum['name']), repr(datum['linked_behavior']))
				_raise(Exception(err))

			z = filter(lambda x: datum['speaking_start'] == x['name'], pony['speak'])
			if len(z) == 0 and datum['speaking_start'] is not None:
				err = "%s: Behavior %s linked to undefined speak %s" % \
			              (ini_file, repr(datum['name']), repr(datum['speaking_start']))
				_raise(Exception(err))

			z = filter(lambda x: datum['speaking_end'] == x['name'], pony['speak'])
			if len(z) == 0 and datum['speaking_end'] is not None:
				err = "%s: Behavior %s linked to undefined speak %s" % \
			              (ini_file, repr(datum['name']), repr(datum['speaking_end']))
				_raise(Exception(err))

			z = filter(lambda x: datum['object_to_follow'] == x['name'], pony['effect'])
			if len(z) == 0 and datum['object_to_follow'] is not None:
				err = "%s: Behavior %s following undefined object %s; is this another Pony?" % \
			              (ini_file, repr(datum['name']), repr(datum['object_to_follow']))
				_raise(Exception(err))

			if datum['probability'] < 0.0 or datum['probability'] > 1.0:
				err = "%s: Behavior %s has a probability of %s; expected 0 <= p <= 1" % \
			              (ini_file, repr(datum['name']), repr(datum['probability']))
				_raise(Exception(err))

			if datum['duration_max'] < 0:
				err = "%s: Behavior %s has a max duration of %s; expected >= 0" % \
			              (ini_file, repr(datum['name']), repr(datum['duration_max']))
				_raise(Exception(err))

			if datum['duration_min'] < 0:
				err = "%s: Behavior %s has a min duration of %s; expected >= 0" % \
			              (ini_file, repr(datum['name']), repr(datum['duration_min']))
				_raise(Exception(err))

			if datum['duration_max'] < datum['duration_min']:
				err = "%s: Behavior %s has a min duration less than max duration" % \
			              (ini_file, repr(datum['name']))
				_raise(Exception(err))

			if datum['movement_speed'] < 0:
				err = "%s: Behavior %s has a movement speed of %s; expected >= 0" % \
			              (ini_file, repr(datum['name']), repr(datum['movement_speed']))
				_raise(Exception(err))

			if datum['x_coord'] < 0.0 or datum['x_coord'] > 1.0:
				err = "%s: Behavior %s has an x coord of %s; expected 0.0 <= x <= 1" % \
			              (ini_file, repr(datum['name']), repr(datum['x_coord']))
				_raise(Exception(err))

			if datum['y_coord'] < 0.0 or datum['y_coord'] > 1.0:
				err = "%s: Behavior %s has a y coord of %s; expected 0.0 <= y <= 1" % \
			              (ini_file, repr(datum['name']), repr(datum['y_coord']))
				_raise(Exception(err))

