# Copyright 2011 Jason Gerecke
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.

"""Module for dealing with Desktop Pony sprites.

This module provides classes for loading Desktop Ponies sprites from
INI files and representing them within the MVC framework that is used.
"""

import random
import time
import math

from ..engine import sprite
from ..engine import task_collection


class Model(sprite.Model):
	"""A Model for sprites which happen to represent ponies.
	
	In addition to housing all the sprite-related data like location
	and speed, the Model contains all the information related to the
	pony, such as its name and behaviors. This can be used by controllers
	to modify the internal state.
	"""

	def __init__(self, universe, info, interact):
		sprite.Model.__init__(self, universe)
		self.name       = info['name']
		self.behaviors  = info['behavior']
		self.effects    = info['effect']
		self.speaks     = info['speak']
		self.categories = info['categories']
		self.interact   = interact

		self.behavior   = None	

	def get_behavior(self, name):
		# Find the new behavior, and set it
		behav_list = [ x for x in self.behaviors if x['name'].upper()==name.upper() ]
		if not behav_list:
			print self.name, "does not have a behavior named", name
			return None
		random.shuffle(behav_list)		
		return behav_list[0]

	def set_behavior(self, name):
		"""Have the Pony begin acting with the specified behavior. """
		b = self.get_behavior(name)
		if not b:
			raise KeyError(name)
		self._set_behavior(b)

	def _set_behavior(self, behavior):
		spoke = False

		# Finish the old behavior by saying the 'speaking_end' lines
		if self.behavior and self.behavior['speaking_end']:
			x = [item for item in self.speaks if item['name'].upper() == self.behavior['speaking_end'].upper()]
			if x:
				self._speak(x[0]['text'])
				spoke = True

		# Set the new behavior, and then set up the animation and motion
		self.behavior = behavior

		if len(self.behavior['movement_angle']) == 0:
			speed = 0
			angle = random.choice([0, 180])
		else:
			speed = self.behavior['movement_speed']
			angle = random.choice(self.behavior['movement_angle'])
		self.set_images(self.behavior['file_animations'])
		self.set_motion(speed=speed, angle=angle)

		# Say any new 'speaking_start' lines, or choose one at random
		if self.behavior and self.behavior['speaking_start']:
			x = [item for item in self.speaks if item['name'].upper() == self.behavior['speaking_start'].upper()]
			if x:
				self._speak(x[0]['text'])
				spoke = True

		if not spoke and random.random() > 0.9:
			x = filter(lambda x: x['skip'] != True, self.speaks)
			if x:
				self._speak(random.choice(x)['text'])
				spoke = True

	def _speak(self, line):
		self._view.Say(self.name + ': "' + line + '"')

	def _get_next_behavior(self):
		if self.behavior and self.behavior['linked_behavior']:
			behavior = self.get_behavior(self.behavior['linked_behavior'])
			if behavior:
				return behavior

		behav_list = [ x for x in self.behaviors if x['probability'] > 0 and not x['skip'] ]
		p_max      = reduce(lambda x, y: x + y['probability'], behav_list, 0)
		p          = random.uniform(0, p_max)

		while p >= 0:
			behavior = behav_list.pop()
			p = p - behavior['probability']

		return behavior


class Controller(sprite.Controller):
	"""A Controller for sprites which happen to represent ponies.
	
	This class acts as a sort of "brain" for ponies. Once started,
	the Pony will begin acting through defined behaviors which specify
	how the Sprite should look, move, and react as the program runs.
	"""

	tasks = task_collection.TaskManager(sleep=0.1)

	def __init__(self, backend, model):
		sprite.Controller.__init__(self, backend, model)
		menu = [ self._behavior_menu() ]
		self.view.set_context_menu(menu)

	def start(self):
		sprite.Controller.start(self)
		self._behaving = True
		self._homing = None
		self._interacting = {}
		self.set_behavior()
		self._interact()

	def stop(self):
		self._behaving = None
		self._homing = None
		self._interacting = None
		sprite.Controller.stop(self)

	def set_behavior(self, behavior=None):
		if behavior is None:
			behavior = self.model._get_next_behavior()
		else:
			behavior = self.model.get_behavior(behavior)
		Controller.tasks.add(self._behave, [behavior], priority=1)

	def _behavior_menu(self):
		menu = []
		for behavior in self.model.behaviors:
			if behavior['skip']:
				continue
			name = behavior['name']
			entry = (name.title(), (self.set_behavior, [name]))
			menu.append(entry)
		return ("Behaviors", menu)

	def _behave(self, behavior):
		if self._behaving is None:
			return
		self._behaving = behavior
		self.model._set_behavior(behavior)
		t_min = self.model.behavior['duration_min']
		t_max = self.model.behavior['duration_max']
		wait = random.uniform(t_min, t_max)

		target_name = self.model.behavior['object_to_follow']
		if target_name is not None:
			targets = self.model.universe.search(self.model, [target_name])
			if len(targets) == 0:
				self._homing = None
			else:
				target = random.choice(targets)
				if target != self._homing:
					self._homing = target
					Controller.tasks.add(self._home, [target])
		else:
			self._homing = None

		Controller.tasks.add(self._behave, [self.model._get_next_behavior()], min_wait = wait)

	def _home(self, target):
		if target is None or target != self._homing:
			return
		source = self.model
		universe = source.universe
		if universe.distance(source, target) > 32: #32 is totally arbitrary.
			theta = math.degrees(universe.angle(source, target))
			source.set_motion(angle=theta)
		Controller.tasks.add(self._home, [target], min_wait = 1)

	def _interact(self):
		if self._interacting is None:
			return

		now = time.time()
		tmp = {}
		for key in self._interacting:
			if self._interacting[key] > now:
				tmp.update({key: self._interacting[key]})
		self._interacting = tmp

		source = self.model
		random.shuffle(source.interact)
		for interact in filter(lambda x: x['name'] not in self._interacting.keys(), source.interact):
			targets = source.universe.search(source, interact['targets'], interact['proximity'])
			targets = filter(lambda x: x != self.model, targets)
			if len(targets) == 0 or \
			   (interact['random_or_all'] == True and len(targets) != len(interact['targets'])):
				continue
			random.shuffle(interact['behaviors'])
			for target in targets:
				target.set_behavior(interact['behaviors'][0])
			self.set_behavior(interact['behaviors'][0])
			self._interacting.update({interact['name']: now + interact['delay']})
		Controller.tasks.add(self._interact, [], min_wait = 1)

